import factory
import pytest
from django.test import RequestFactory
from rest_framework.reverse import reverse
from wbaccounting.models import BookingEntry
from wbaccounting.serializers import (
    BookingEntryModelSerializer,
    BookingEntryRepresentationSerializer,
)


@pytest.mark.django_db
class TestBookingEntryModelSerializer:
    def test_serialize(self, booking_entry: BookingEntry):
        serializer = BookingEntryModelSerializer(booking_entry)
        assert isinstance(serializer.data, dict)

    def test_serialize_contains_keys(self, booking_entry: BookingEntry):
        serializer = BookingEntryModelSerializer(booking_entry)
        assert (
            "id",
            "title",
            "booking_date",
            "due_date",
            "payment_date",
            "reference_date",
            "net_value",
            "gross_value",
            "vat",
            "invoice_net_value",
            "invoice_gross_value",
            "invoice_fx_rate",
            "currency",
            "_currency",
            "counterparty",
            "_counterparty",
            "invoice",
            "_invoice",
            "invoice_currency",
            "_additional_resources",
            "_buttons",
        ) == tuple(
            serializer.data.keys()
        )  # type: ignore

    @pytest.mark.parametrize("decorator", ["net_value", "gross_value"])
    def test_serialize_decorators(self, booking_entry: BookingEntry, rf: RequestFactory, decorator: str):
        serializer = BookingEntryModelSerializer(booking_entry)
        _, representation = serializer[decorator].get_representation(rf.get("/"), decorator)  # type: ignore
        assert representation["decorators"] == [{"position": "left", "type": "text", "value": "{{_currency.symbol}}"}]

    def test_deserialize(self, booking_entry_factory, entry, currency, invoice):
        data = factory.build(dict, FACTORY_CLASS=booking_entry_factory)
        data["currency"] = currency.pk
        data["counterparty"] = entry.pk
        data["invoice"] = invoice.pk
        serializer = BookingEntryModelSerializer(data=data)
        assert serializer.is_valid()


@pytest.mark.django_db
class TestBookingEntryRepresentationSerializer:
    def test_serialize(self, booking_entry: BookingEntry):
        serializer = BookingEntryRepresentationSerializer(booking_entry)
        assert serializer.data == {
            "id": booking_entry.pk,
            "title": booking_entry.title,
            "_detail": reverse("wbaccounting:bookingentry-detail", args=[booking_entry.pk]),
        }



from django.db.models import QuerySet
from wbcore import viewsets

from wbaccounting.models import BookingEntry
from wbaccounting.serializers import (
    BookingEntryModelSerializer,
    BookingEntryRepresentationSerializer,
)
from wbaccounting.viewsets.buttons import BookingEntryButtonConfig
from wbaccounting.viewsets.display import BookingEntryDisplayConfig
from wbaccounting.viewsets.titles import BookingEntryTitleConfig


class BookingEntryRepresentationViewSet(viewsets.RepresentationViewSet):
    search_fields = ["counterparty__computed_str", "title", "currency__key"]
    ordering_fields = ["title"]
    ordering = ["title"]
    serializer_class = BookingEntryRepresentationSerializer
    queryset = BookingEntry.objects.all()

    def get_queryset(self):
        return BookingEntry.objects.filter_for_user(self.request.user)


class BookingEntryModelViewSet(viewsets.ModelViewSet):
    filterset_fields = {
        "counterparty": ["exact"],
        "booking_date": ["gte", "exact", "lte"],
        "reference_date": ["gte", "exact", "lte"],
    }
    search_fields = ["counterparty__computed_str", "title", "currency__key"]
    ordering_fields = (
        "counterparty__computed_str",
        "booking_date",
        "gross_value",
        "net_value",
        "reference_date",
    )
    ordering = ["-booking_date"]

    serializer_class = BookingEntryModelSerializer
    queryset = BookingEntry.objects.all()

    button_config_class = BookingEntryButtonConfig
    display_config_class = BookingEntryDisplayConfig
    title_config_class = BookingEntryTitleConfig

    def get_queryset(self) -> QuerySet[BookingEntry]:
        booking_entries = BookingEntry.objects.filter_for_user(self.request.user)

        if eai_id := self.kwargs.get("entry_accounting_information_id", None):
            booking_entries = booking_entries.filter(counterparty__entry_accounting_information__id=eai_id)

        if invoice_id := self.kwargs.get("invoice_id", None):
            booking_entries = booking_entries.filter(invoice_id=invoice_id)

        return booking_entries.select_related(
            "counterparty",
            "invoice",
            "currency",
            "invoice__invoice_currency",
        )



from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class BookingEntryDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        _counterparty_fields = [dp.Field(key="counterparty", label="Counterparty")]
        _common_fields = [
            dp.Field(key="booking_date", label="Book. Date", width=100),
            dp.Field(key="reference_date", label="Ref. Date", width=100),
            dp.Field(key="payment_date", label="Payment. Date", width=100),
            dp.Field(key="title", label="Title", width=325),
            dp.Field(key="vat", label="VAT", width=100),
            dp.Field(key="gross_value", label="Gross", width=100),
            dp.Field(key="net_value", label="Net", width=100),
        ]
        _invoice_fields = [dp.Field(key="invoice", label="Invoice", width=325)]

        if "invoice_id" in self.view.kwargs:
            return dp.ListDisplay(fields=_common_fields)

        if "entry_accounting_information_id" in self.view.kwargs:
            return dp.ListDisplay(fields=[*_common_fields, *_invoice_fields])

        return dp.ListDisplay(fields=[*_counterparty_fields, *_common_fields, *_invoice_fields])

    def get_instance_display(self) -> dp.Display:
        return dp.Display(
            pages=[
                dp.Page(
                    layouts={
                        dp.default(): dp.Layout(
                            grid_template_areas=[
                                ["title", "title", "title", "counterparty", ".", "booking_date"],
                                ["currency", "vat", "gross_value", "net_value", ".", "reference_date"],
                                [
                                    "invoice_currency",
                                    "invoice_fx_rate",
                                    "invoice_gross_value",
                                    "invoice_net_value",
                                    ".",
                                    "due_date",
                                ],
                                ["invoice", "invoice", "invoice", "invoice", ".", "payment_date"],
                            ],
                            grid_template_columns=[
                                "100px",
                                "100px",
                                "200px",
                                "200px",
                                "25px",
                                "150px",
                            ],
                        )
                    }
                )
            ]
        )



import pytest
from dynamic_preferences.registries import global_preferences_registry
from wbaccounting.models import BookingEntry, EntryAccountingInformation
from wbaccounting.models.entry_accounting_information import (
    default_currency,
    default_email_body,
)
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import Currency


@pytest.mark.django_db
class TestEntryAccountingInformation:
    def test_str(self, entry_accounting_information: EntryAccountingInformation):
        assert str(entry_accounting_information) == f"Counterparty: {entry_accounting_information.entry.computed_str}"

    @pytest.mark.parametrize(
        "method,return_value",
        [
            ("get_endpoint_basename", "wbaccounting:entryaccountinginformation"),
            ("get_representation_value_key", "id"),
            ("get_representation_label_key", "{{entry_repr}}"),
            ("get_representation_endpoint", "wbaccounting:entryaccountinginformationrepresentation-list"),
        ],
    )
    def test_wbmodel_methods(self, method: str, return_value: str):
        assert getattr(EntryAccountingInformation, method)() == return_value

    @pytest.mark.parametrize("user__is_superuser", [False])
    def test_filter_for_user_no_superuser(self, entry_accounting_information: EntryAccountingInformation, user: User):
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information not in entry_accounting_information_list

    @pytest.mark.parametrize("user__is_superuser", [True])
    def test_filter_for_user_superuser(self, entry_accounting_information: EntryAccountingInformation, user: User):
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information in entry_accounting_information_list

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [False])
    def test_filter_for_user_public_counterparty(
        self, entry_accounting_information: EntryAccountingInformation, user: User
    ):
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information in entry_accounting_information_list

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty(
        self, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information not in entry_accounting_information_list

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty_with_exempt(
        self, booking_entry: BookingEntry, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        entry_accounting_information.exempt_users.add(user)
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information in entry_accounting_information_list

    def test_default_email_body(self):
        assert default_email_body() == ""

    def test_custom_default_email_body(self):
        global_preferences_registry.manager()["wbaccounting__invoice_email_body"] = "Custom Body"
        assert default_email_body() == "Custom Body"

    def test_default_currency(self):
        assert default_currency() is None

    def test_custom_default_currency(self, currency: Currency):
        global_preferences_registry.manager()[
            "wbaccounting__default_entry_account_information_currency_key"
        ] = currency.key
        assert default_currency().pk is currency.pk



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from wbaccounting.viewsets import (
    BookingEntryModelViewSet,
    BookingEntryRepresentationViewSet,
    ConsolidatedInvoiceViewSet,
    EntryAccountingInformationModelViewSet,
    EntryAccountingInformationRepresentationViewSet,
    FutureCashFlowPandasAPIViewSet,
    FutureCashFlowTransactionsPandasAPIViewSet,
    InvoiceModelViewSet,
    InvoiceRepresentationViewSet,
    InvoiceTypeModelViewSet,
    InvoiceTypeRepresentationViewSet,
    TransactionModelViewSet,
    TransactionRepresentationViewSet,
)

router = WBCoreRouter()
router.register(r"bookingentry", BookingEntryModelViewSet, basename="bookingentry")
router.register(
    r"bookingentryrepresentation", BookingEntryRepresentationViewSet, basename="bookingentryrepresentation"
)
router.register(r"invoice", InvoiceModelViewSet, basename="invoice")
router.register(r"invoicerepresentation", InvoiceRepresentationViewSet, basename="invoicerepresentation")
router.register(r"transaction", TransactionModelViewSet, basename="transaction")
router.register(r"transactionrepresentation", TransactionRepresentationViewSet, basename="transactionrepresentation")
router.register(r"consolidated-invoice", ConsolidatedInvoiceViewSet, basename="consolidated-invoice")
router.register(r"invoicetype", InvoiceTypeModelViewSet, basename="invoicetype")
router.register(r"invoicetyperepresentation", InvoiceTypeRepresentationViewSet, basename="invoicetyperepresentation")
router.register(
    r"entryaccountinginformationrepresentation",
    EntryAccountingInformationRepresentationViewSet,
    basename="entryaccountinginformationrepresentation",
)
router.register(
    r"entryaccountinginformation", EntryAccountingInformationModelViewSet, basename="entryaccountinginformation"
)
router.register(r"futurecashflow", FutureCashFlowPandasAPIViewSet, basename="futurecashflow")
router.register(
    r"futurecashflowtransaction", FutureCashFlowTransactionsPandasAPIViewSet, basename="futurecashflowtransaction"
)

entry_router = WBCoreRouter()

invoice_router = WBCoreRouter()
invoice_router.register(r"bookingentry", BookingEntryModelViewSet, basename="invoice-bookingentry")

entry_accounting_information_router = WBCoreRouter()
entry_accounting_information_router.register(
    r"invoice", InvoiceModelViewSet, basename="entryaccountinginformation-invoice"
)
entry_accounting_information_router.register(
    r"bookingentry", BookingEntryModelViewSet, basename="entryaccountinginformation-bookingentry"
)

booking_entry_router = WBCoreRouter()

urlpatterns = [
    path("", include(router.urls)),
    path("entry/<int:entry_id>/", include(entry_router.urls)),
    path("invoice/<int:invoice_id>/", include(invoice_router.urls)),
    path("bookingentry/<int:booking_entry_id>/", include(booking_entry_router.urls)),
    path(
        "entryaccountinginformation/<int:entry_accounting_information_id>/",
        include(entry_accounting_information_router.urls),
    ),
]



from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class BookingEntryButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self):
        return {
            buttons.WidgetButton(
                key="appended_booking_entries", label="Appended Booking Entries", icon=WBIcon.NOTEBOOK.icon
            )
        }

    def get_custom_list_instance_buttons(self):
        return self.get_custom_instance_buttons()



import pytest
from wbaccounting.models import InvoiceType


@pytest.mark.django_db
class TestInvoiceType:
    def test_str(self, invoice_type: InvoiceType):
        assert str(invoice_type) == invoice_type.name

    @pytest.mark.parametrize(
        "method,return_value",
        [
            ("get_endpoint_basename", "wbaccounting:invoicetype"),
            ("get_representation_value_key", "id"),
            ("get_representation_label_key", "{{name}}"),
            ("get_representation_endpoint", "wbaccounting:invoicetyperepresentation-list"),
        ],
    )
    def test_wbmodel_methods(self, method: str, return_value: str):
        assert getattr(InvoiceType, method)() == return_value




from django.template import Context, Template
from reportlab.lib import colors, utils
from reportlab.lib.enums import TA_CENTER, TA_RIGHT
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle, getSampleStyleSheet
from reportlab.lib.units import cm
from reportlab.platypus import Image, Paragraph, Table, TableStyle

from wbaccounting.dynamic_preferences_registry import format_invoice_number


def get_styles():
    styles = getSampleStyleSheet()
    styles.add(ParagraphStyle(name="Headline", fontName="Helvetica", fontSize=24, leading=34))
    styles.add(ParagraphStyle(name="Default", fontName="Helvetica", fontSize=10, leading=12))
    styles.add(ParagraphStyle(name="Default-Table-Header", fontName="Helvetica", fontSize=14, leading=16))
    styles.add(
        ParagraphStyle(
            name="Default-Table-Header-Right", fontName="Helvetica", fontSize=14, leading=16, alignment=TA_RIGHT
        )
    )
    styles.add(ParagraphStyle(name="Default-Right", fontName="Helvetica", fontSize=10, leading=12, alignment=TA_RIGHT))
    styles.add(
        ParagraphStyle(
            name="Default-Right-Linebreak",
            fontName="Helvetica",
            fontSize=10,
            leading=12,
            alignment=TA_RIGHT,
            spaceAfter=12,
        )
    )
    styles.add(
        ParagraphStyle(name="Default-Center", fontName="Helvetica", fontSize=10, leading=12, alignment=TA_CENTER)
    )
    styles.add(ParagraphStyle(name="Default-Linebreak", fontName="Helvetica", fontSize=10, leading=12, spaceAfter=12))
    styles.add(
        ParagraphStyle(name="Default-Linebreak-Before", fontName="Helvetica", fontSize=10, leading=12, spaceBefore=12)
    )
    styles.add(
        ParagraphStyle(name="Default-Double-Linebreak", fontName="Helvetica", fontSize=10, leading=12, spaceAfter=24)
    )
    styles.add(
        ParagraphStyle(
            name="Default-Double-Linebreak-With-Before",
            fontName="Helvetica",
            fontSize=10,
            leading=12,
            spaceAfter=24,
            spaceBefore=24,
        )
    )
    (styles.add(ParagraphStyle(name="Normal-Right", alignment=TA_RIGHT)),)
    (styles.add(ParagraphStyle(name="Annotation", fontName="Helvetica", fontSize=8, leading=10)),)
    (
        styles.add(
            ParagraphStyle(name="Annotation-Right", fontName="Helvetica", fontSize=8, leading=10, alignment=TA_RIGHT)
        ),
    )
    return styles


# TODO to be removed
# def get_taxes(invoice, party, counterparty):
#     vat = 0
#     social_charges = 0
#     if invoice.favour == 'COUNTERPARTY':
#         vat = counterparty.accounting_information.vat if counterparty.accounting_information.vat else 0
#         social_charges = counterparty.accounting_information.social_charges if counterparty.accounting_information.social_charges else 0
#     else:
#         vat = party.vat if party is not None and party.vat else 0
#         social_charges = party.accounting_information.social_charges if party is not None and party.accounting_information.social_charges else 0
#     if invoice.vat or invoice.vat == 0:
#         vat = invoice.vat
#     if invoice.social_charges or invoice.social_charges == 0:
#         social_charges = invoice.social_charges
#     return vat, social_charges

# def banking_block(entry, invoice, styles, banking=None):
#     elements = list()
#     if banking is None:
#         bankings = entry.banking.all() if entry else None
#         if bankings:
#             if bankings.filter(currency=invoice.base_currency).exists():
#                 banking = bankings.filter(currency=invoice.base_currency).first()
#             elif bankings.filter(primary=True).exists():
#                 banking = bankings.filter(primary=True).first()
#         else:
#             return elements
#     elements.append(Paragraph('{}'.format(banking.institute), styles['Default-Linebreak-Before']))
#     if banking.institute_additional:
#         elements.append(Paragraph('{}'.format(banking.institute_additional), styles['Default']))
#     elements.append(Paragraph('IBAN: {}'.format(banking.iban), styles['Default']))
#     if banking.swift_bic:
#         elements.append(Paragraph('SWIFT: {}'.format(banking.swift_bic), styles['Default']))
#     return elements


def address_block(entry, invoice, styles, right=False, address=None, tax_id=None):
    elements = list()
    if address is None:
        addresses = entry.addresses.all() if entry is not None else None
        if addresses:
            if addresses.filter(primary=True).exists():
                address = addresses.filter(primary=True).first()
            elif addresses.count() > 0:
                address = addresses.first()
        else:
            return elements

    if right:
        default_style = styles["Default-Right"]
        default_linebreak_style = styles["Default-Right-Linebreak"]
    else:
        default_style = styles["Default"]
        default_linebreak_style = styles["Default-Linebreak"]

    casted_entry = entry.get_casted_entry()
    if entry.is_company:
        elements.append(Paragraph("<b>{0.name}</b>".format(casted_entry), default_style))
    else:
        elements.append(Paragraph("<b>{0.first_name} {0.last_name}</b>".format(casted_entry), default_style))

    elements.append(Paragraph(address.street, default_style))
    if address.street_additional is not None and address.street_additional != "":
        elements.append(Paragraph(address.street_additional, default_style))
    elements.append(Paragraph("{} {}".format(address.zip, address.geography_city), default_style))
    elements.append(Paragraph(address.geography_city.parent.parent.name, default_linebreak_style))

    if tax_id and entry.entry_accounting_information.tax_id:
        elements.append(Paragraph(f"VAT: {entry.entry_accounting_information.tax_id}", default_linebreak_style))

    return elements


def city_and_date_block(invoice, styles, entry=None, city=None, date=None):
    elements = list()
    if date is None:
        date = invoice.invoice_date
    if city is None:
        primary_address = entry.primary_address_contact() if entry else None
        if primary_address:
            city = primary_address.geography_city
    if city:
        elements.append(Paragraph("{}, {:%d.%m.%Y}".format(city, date), styles["Default-Right-Linebreak"]))
    else:
        elements.append(Paragraph("{:%d.%m.%Y}".format(date), styles["Default-Right-Linebreak"]))
    return elements


def logo_block(image):
    # print(image)
    reportlab_image = utils.ImageReader(image)
    width, height = reportlab_image.getSize()
    ratio = float(width) / float(height)
    image = Image(image, width=30 * ratio, height=30)
    return image, ratio


def signature_block(signees, styles):
    elements = list()
    if signees:
        elements.append(Paragraph("Best Regards,", styles["Default-Double-Linebreak-With-Before"]))
        for signee in signees:
            if signee is not None and signee.signature.name is not None and signee.signature.name != "":
                reportlab_image = utils.ImageReader(signee.signature)
                width, height = reportlab_image.getSize()
                ratio = float(width) / float(height)

                image = Image(signee.signature, width=50 * ratio, height=50)
                image.hAlign = "LEFT"
                elements.append(image)
                elements.append(Paragraph("{}".format(signee.computed_str), styles["Default"]))
    return elements


def add_text_block_with_context(text, invoice, styles):
    elements = list()
    if text:
        elements.append(Paragraph("", styles["Default-Linebreak"]))
        elements.append(Paragraph(Template(text).render(Context(invoice.get_context())), styles["Default-Linebreak"]))
    return elements


def add_transfer_block(invoice, receiver, styles, total_amount_gross):
    elements = list()

    if invoice.is_counterparty_invoice:
        text = f"The amount of <b>{invoice.invoice_currency} {format_invoice_number(total_amount_gross)}</b> will be transfered to the following bank account:"
    else:
        text = f"Please transfer <b>{invoice.invoice_currency} {format_invoice_number(total_amount_gross)}</b> to the following bank account:"

    elements.append(Paragraph(text, styles["Default-Linebreak-Before"]))

    bank_accounts = receiver.banking.all()
    if bank_accounts.exists():
        try:
            bank_account = bank_accounts.get(currency=invoice.invoice_currency)
        except Exception:
            bank_account = bank_accounts.first()

        elements.append(Paragraph(bank_account.institute, styles["Default"]))

        if bank_account.institute_additional:
            elements.append(Paragraph(bank_account.institute_additional, styles["Default"]))

        if bank_account.iban and bank_account.iban != "":
            elements.append(Paragraph(f"IBAN: {bank_account.iban}", styles["Default"]))
        if bank_account.swift_bic:
            elements.append(Paragraph(f"SWIFT: {bank_account.swift_bic}", styles["Default"]))
        if bank_account.additional_information and bank_account.additional_information != "":
            elements.append(Paragraph(bank_account.additional_information, styles["Default"]))

    return elements


def get_gross_and_net_value(invoice):
    multiplier = 1 if not invoice.is_counterparty_invoice else -1

    total_gross = 0
    total_net = 0
    for booking_entry in invoice.booking_entries.all():
        if booking_entry.currency != invoice.invoice_currency:
            conversion_rate = booking_entry.currency.convert(booking_entry.booking_date, invoice.invoice_currency)
            total_net += multiplier * booking_entry.net_value * conversion_rate
            total_gross += multiplier * booking_entry.gross_value * conversion_rate
        else:
            total_net += multiplier * booking_entry.net_value
            total_gross += multiplier * booking_entry.gross_value

    return total_gross, total_net


def add_booking_entries(invoice, styles, inverse=False):
    multiplier = 1 if not invoice.is_counterparty_invoice else -1

    tables = list()
    for vat in invoice.booking_entries.all().distinct("vat").values_list("vat", flat=True):
        table_data = list()

        table_data.append([Paragraph("", styles["Default"])] * 2)
        table_data.append(
            [
                Paragraph("Title", styles["Default-Table-Header"]),
                Paragraph("Amount", styles["Default-Table-Header-Right"]),
            ]
        )

        for booking_entry in invoice.booking_entries.filter(vat=vat).order_by("title"):
            title_col = [Paragraph(f"{booking_entry.title}", styles["Default"])]
            amount_col = list()

            if booking_entry.currency != invoice.invoice_currency:
                conversion_rate = booking_entry.currency.convert(booking_entry.booking_date, invoice.invoice_currency)

                title_col.append(
                    Paragraph(
                        f"<i>1.00 {booking_entry.currency.key} = {conversion_rate:.4f} {invoice.invoice_currency.key}</i>",
                        styles["Annotation"],
                    )
                )

                amount_col.append(
                    Paragraph(
                        f"<i>{format_invoice_number(multiplier * conversion_rate * booking_entry.gross_value)} {invoice.invoice_currency.key}</i>",
                        styles["Default-Right"],
                    )
                )
                amount_col.append(
                    Paragraph(
                        f"{format_invoice_number(multiplier * booking_entry.gross_value)} {booking_entry.currency}",
                        styles["Annotation-Right"],
                    )
                )
            else:
                amount_col.append(
                    Paragraph(
                        f"{format_invoice_number(multiplier * booking_entry.gross_value)} {booking_entry.currency}",
                        styles["Default-Right"],
                    )
                )

            if (
                (params := booking_entry.parameters)
                and (from_date := params.get("from_date"))
                and (to_date := params.get("to_date"))
            ):
                title_col.append(
                    Paragraph(
                        f"<i>{from_date} - {to_date}</i>",
                        styles["Annotation"],
                    )
                )

            table_data.append([title_col, amount_col])

        total_gross, total_net = get_gross_and_net_value(invoice)
        table_data.append([Paragraph("", styles["Default"])] * 2)
        table_data.append(
            [
                [
                    Paragraph("<strong>Total Amount</strong>", styles["Default"]),
                    Paragraph(f"<i>Vat: {vat*100:.2f}%</i>", styles["Annotation"]),
                    Paragraph(f"Total Amount (excl. {vat*100:.2f}% VAT)", styles["Annotation"]),
                ],
                [
                    Paragraph(
                        f"<strong>{format_invoice_number(total_gross)} {invoice.invoice_currency}</strong>",
                        styles["Default-Right"],
                    ),
                    Paragraph(
                        f"<i>{format_invoice_number(total_gross - total_net)} {invoice.invoice_currency}</i>",
                        styles["Annotation-Right"],
                    ),
                    Paragraph(
                        f"{format_invoice_number(total_net)} {invoice.invoice_currency}",
                        styles["Annotation-Right"],
                    ),
                ],
            ]
        )
        table = Table(table_data, colWidths=[A4[0] - 6.4 * cm, 3.5 * cm])
        table.setStyle(
            TableStyle(
                [
                    ("BACKGROUND", (0, 0), (-1, 0), colors.HexColor("#f37200")),
                    ("BACKGROUND", (0, 1), (-1, 1), colors.HexColor("#f38a32")),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                ]
            )
        )
        tables.append(table)

    return tables



from wbcore.metadata.configs.titles import TitleViewConfig


class InvoiceTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Invoice {{title}}"

    def get_list_title(self):
        return "Invoices"

    def get_create_title(self):
        return "New Invoice"


class ConsolidatedInvoiceTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Consolidated Invoice"

    def get_list_title(self):
        return "Consolidated Invoices"

    def get_create_title(self):
        return "New Consolidated Invoice"



from contextlib import suppress
from datetime import date

from django.db import models
from django.db.models import Q, QuerySet
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.module_loading import import_string
from dynamic_preferences.registries import global_preferences_registry as gpr
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import Currency

from wbaccounting.models.model_tasks import generate_booking_entries_as_task


def default_email_body() -> str:
    with suppress(Exception):
        return gpr.manager()["wbaccounting__invoice_email_body"]

    return ""


def default_currency() -> Currency | None:
    with suppress(Exception):
        return Currency.objects.get(key=gpr.manager()["wbaccounting__default_entry_account_information_currency_key"])


class EntryAccountingInformationDefaultQuerySet(QuerySet):
    def filter_for_user(self, user: User) -> QuerySet:
        """
        Filters entry accounting information based on if the current user is allowed to see it.

        Args:
            user (User): The user for whom entry accounting information need to be filtered.

        Returns:
            QuerySet: A filtered queryset.
        """

        if user.is_superuser or user.has_perm("wbaccounting.administrate_invoice"):
            return self

        if not user.has_perm("wbaccounting.view_entryaccountinginformation"):
            return self.none()

        return self.filter(Q(counterparty_is_private=False) | Q(exempt_users=user))


class EntryAccountingInformation(models.Model):
    # Link to Entry
    entry = models.OneToOneField(
        "directory.Entry",
        on_delete=models.CASCADE,
        related_name="entry_accounting_information",
        verbose_name="Linked Counterparty",
    )

    # Tax Information
    tax_id = models.CharField(max_length=512, blank=True, null=True, verbose_name="Tax ID")
    vat = models.FloatField(blank=True, null=True, verbose_name="VAT")

    # Invoice Information
    default_currency = models.ForeignKey(
        "currency.Currency",
        related_name="entry_accounting_informations",
        default=default_currency,
        on_delete=models.PROTECT,
        verbose_name="Default Currency",
        blank=True,
        null=True,
    )

    email_to = models.ManyToManyField(
        "directory.EmailContact", related_name="entry_accounting_informations_to", blank=True, verbose_name="To"
    )

    email_cc = models.ManyToManyField(
        "directory.EmailContact", related_name="entry_accounting_informations_cc", blank=True, verbose_name="CC"
    )
    email_bcc = models.ManyToManyField(
        "directory.EmailContact", related_name="entry_accounting_informations_bcc", blank=True, verbose_name="BCC"
    )

    email_subject = models.CharField(default="{{invoice.title}}", max_length=1024, verbose_name="Subject")
    email_body = models.TextField(default=default_email_body, verbose_name="Body")
    send_mail = models.BooleanField(default=True, verbose_name="Send Mail")

    counterparty_is_private = models.BooleanField(
        default=False,
        verbose_name="Counterparty Is Private",
        help_text="Hides all of the counterparty's invoices from non-eligible users",
    )
    exempt_users = models.ManyToManyField(
        to="authentication.User",
        verbose_name="Exempt Users",
        help_text="Exclusion list of users who are able to see private invoices for the counterparty",
        related_name="private_accounting_information",
        blank=True,
    )
    booking_entry_generator = models.CharField(max_length=256, null=True, blank=True)
    default_invoice_type = models.ForeignKey(
        to="wbaccounting.InvoiceType",
        related_name="booking_entries",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Default Invoice Type",
        help_text="When invoicinging outstanding booking entries, this invoice type will be assigned to the corresponding invoice",
    )

    external_invoice_users = models.ManyToManyField(
        to="authentication.User",
        verbose_name="External User",
        help_text="External users who are able to see the invoices generated for this counterparty",
        related_name="external_accounting_information",
        blank=True,
    )

    def get_booking_entry_generator(self):
        with suppress(ImportError):
            return import_string(self.booking_entry_generator or "")

    objects = EntryAccountingInformationDefaultQuerySet.as_manager()

    class Meta:
        verbose_name = "Counterparty"
        verbose_name_plural = "Counterparties"

    def __str__(self):
        return f"Counterparty: {self.entry.computed_str}"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbaccounting:entryaccountinginformationrepresentation-list"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbaccounting:entryaccountinginformation"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{entry_repr}}"

    def generate_booking_entries(self, from_date: date, to_date: date):
        generate_booking_entries_as_task.delay(  # type: ignore
            self.booking_entry_generator or "", from_date, to_date, self.entry.id
        )


@receiver(post_save, sender=EntryAccountingInformation)
def post_save_entry(sender, instance, **kwargs):
    """If the EAI does not have any email_to, then we add the entries primary email address to it (if it exists)"""
    if not instance.email_to.exists() and (email := instance.entry.primary_email_contact()):
        instance.email_to.add(email)



from django.dispatch import receiver
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.directory.serializers import EntryRepresentationSerializer
from wbcore.contrib.directory.viewsets import (
    CompanyModelViewSet,
    EntryModelViewSet,
    PersonModelViewSet,
)
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.signals.instance_buttons import add_instance_button
from wbcore.utils.date import current_quarter_date_end, current_quarter_date_start


class StartEndParametersSerializer(serializers.Serializer):
    start = serializers.DateField(default=current_quarter_date_start(), label="Start")
    end = serializers.DateField(default=current_quarter_date_end(), label="End")


class CounterpartiesSerializer(serializers.Serializer):
    counterparties = serializers.PrimaryKeyRelatedField(many=True)
    _counterparties = EntryRepresentationSerializer(many=True, source="counterparties")


class StartEndParametersWithCounterpartiesSerializer(CounterpartiesSerializer, StartEndParametersSerializer): ...


class EntryAccountingInformationButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self) -> set:
        if not self.instance:
            return {
                buttons.ActionButton(
                    method=RequestType.POST,
                    identifiers=("wbaccounting:bookingentry",),
                    endpoint=reverse(
                        "wbaccounting:entryaccountinginformation-generate-booking-entries-for-counterparties",
                        request=self.request,
                    ),
                    label="Generate Booking Entries",
                    description_fields="""<p>Generate Booking Entries between {{start}} and {{end}}?</p>""",
                    action_label="Generate Booking Entries",
                    title="Generate Booking Entries",
                    serializer=StartEndParametersWithCounterpartiesSerializer,
                    instance_display=create_simple_display([["start"], ["end"], ["counterparties"]]),
                ),
                buttons.ActionButton(
                    method=RequestType.POST,
                    identifiers=("wbaccounting:invoice",),
                    endpoint=reverse(
                        "wbaccounting:entryaccountinginformation-generate-invoices-for-counterparties",
                        request=self.request,
                    ),
                    label="Invoice outstanding Bookings",
                    description_fields="""<p>Invoice outstanding bookings? If you don't supply counterparties, all counterparties will be considered</p>""",
                    action_label="Invoice outstanding Bookings",
                    title="Invoice outstanding Bookings",
                    serializer=CounterpartiesSerializer,
                    instance_display=create_simple_display([["counterparties"]]),
                ),
            }
        return set()

    def get_custom_instance_buttons(self) -> set:
        return {
            buttons.ActionButton(
                method=RequestType.POST,
                identifiers=("wbaccounting:bookingentry",),
                key="generate_booking_entries",
                label="Generate Booking Entries",
                description_fields="""<p>Generate Booking Entries between {{start}} and {{end}}?</p>""",
                action_label="Generate Booking Entries",
                title="Generate Booking Entries",
                serializer=StartEndParametersSerializer,
                instance_display=create_simple_display([["start"], ["end"]]),
            ),
            buttons.ActionButton(
                method=RequestType.POST,
                identifiers=("wbaccounting:invoice",),
                key="invoice_booking_entries",
                label="Invoice Booking Entries",
                description_fields="<p>Do you want to invoice all outstanding booking entries?</p>",
                action_label="Invoice Booking Entries",
                title="Invoice Booking Entries",
            ),
        }


@receiver(add_instance_button, sender=PersonModelViewSet)
@receiver(add_instance_button, sender=EntryModelViewSet)
@receiver(add_instance_button, sender=CompanyModelViewSet)
def entry_adding_instance_buttons(sender, many, *args, **kwargs):
    if not many:
        return buttons.WidgetButton(key="accounting-information", label="Accounting", icon=WBIcon.MAIL_OPEN.icon)



from typing import Optional

from rest_framework.reverse import reverse
from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbaccounting.models import Invoice

COLOR_DRAFT = "rgb(255,249,196)"
COLOR_SUBMITTED = "rgb(51, 204, 255)"
COLOR_SENT = "rgb(255, 153, 102)"
COLOR_PAID = "rgb(214, 229, 145)"


class InvoiceDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="status", label="Status"),
                dp.Field(key="type", label="Type"),
                dp.Field(key="counterparty", label="Counterparty"),
                dp.Field(key="title", label="Title"),
                dp.Field(key="invoice_date", label="Invoice Date"),
                dp.Field(key="reference_date", label="Reference Date"),
                dp.Field(key="gross_value", label="Gross Value"),
                dp.Field(key="net_value", label="Net Value"),
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=COLOR_DRAFT,
                            label=Invoice.Status.DRAFT.label,
                            value=Invoice.Status.DRAFT.name,
                        ),
                        dp.LegendItem(
                            icon=COLOR_SUBMITTED,
                            label=Invoice.Status.SUBMITTED.label,
                            value=Invoice.Status.SUBMITTED.name,
                        ),
                        dp.LegendItem(
                            icon=WBColor.BLUE_LIGHT.value,
                            label=Invoice.Status.APPROVED.label,
                            value=Invoice.Status.APPROVED.name,
                        ),
                        dp.LegendItem(
                            icon=COLOR_SENT,
                            label=Invoice.Status.SENT.label,
                            value=Invoice.Status.SENT.name,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=Invoice.Status.PAID.label,
                            value=Invoice.Status.PAID.name,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREY.value,
                            label=Invoice.Status.CANCELLED.label,
                            value=Invoice.Status.CANCELLED.name,
                        ),
                    ],
                )
            ],
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": COLOR_DRAFT},
                            condition=("==", Invoice.Status.DRAFT.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": COLOR_SUBMITTED},
                            condition=("==", Invoice.Status.SUBMITTED.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=("==", Invoice.Status.APPROVED.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": COLOR_SENT},
                            condition=("==", Invoice.Status.SENT.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", Invoice.Status.PAID.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREY.value},
                            condition=("==", Invoice.Status.CANCELLED.name),
                        ),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "status")],
                ["title", "invoice_type"],
                ["counterparty", "reference_date"],
                ["invoice_date", "invoice_currency"],
                ["text_above", "text_below"],
                [repeat_field(2, "booking_entry_section")],
            ],
            [
                create_simple_section(
                    "booking_entry_section", "Booking Entries", [["bookingentries"]], "bookingentries", collapsed=True
                )
            ],
        )


class ConsolidatedInvoiceDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="group", label="Group", pinned="left"),
                dp.Field(key="value", label="Value"),
            ],
            tree=True,
            tree_group_field="group",
            tree_group_level_options=[
                dp.TreeGroupLevelOption(
                    filter_depth=None,  # We want lookup to be concatenated
                    filter_key="lookup",
                    list_endpoint=reverse(
                        "wbaccounting:consolidated-invoice-list",
                        args=[],
                        request=self.request,
                    ),
                )
            ],
            legends=[
                dp.Legend(
                    key="status",
                    label="Lowest Status",
                    items=[
                        dp.LegendItem(
                            icon=COLOR_DRAFT,
                            label=Invoice.Status.DRAFT.label,
                            value=Invoice.Status.DRAFT.name,
                        ),
                        dp.LegendItem(
                            icon=COLOR_SUBMITTED,
                            label=Invoice.Status.SUBMITTED.label,
                            value=Invoice.Status.SUBMITTED.name,
                        ),
                        dp.LegendItem(
                            icon=WBColor.BLUE_LIGHT.value,
                            label=Invoice.Status.APPROVED.label,
                            value=Invoice.Status.APPROVED.name,
                        ),
                        dp.LegendItem(
                            icon=COLOR_SENT,
                            label=Invoice.Status.SENT.label,
                            value=Invoice.Status.SENT.name,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=Invoice.Status.PAID.label,
                            value=Invoice.Status.PAID.name,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREY.value,
                            label=Invoice.Status.CANCELLED.label,
                            value=Invoice.Status.CANCELLED.name,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="num_paid",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": COLOR_PAID},
                            condition=(">", 0),
                        )
                    ],
                ),
                dp.Formatting(
                    column="num_sent",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": COLOR_SENT},
                            condition=(">", 0),
                        )
                    ],
                ),
                dp.Formatting(
                    column="num_submitted",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": COLOR_SUBMITTED},
                            condition=(">", 0),
                        )
                    ],
                ),
                dp.Formatting(
                    column="num_draft",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": COLOR_DRAFT},
                            condition=(">", 0),
                        )
                    ],
                ),
            ],
        )



from wbcore import serializers

from wbaccounting.models import InvoiceType


class InvoiceTypeModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = InvoiceType
        fields = ("id", "name", "processor")


class InvoiceTypeRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbaccounting:invoicetype-detail")

    class Meta:
        model = InvoiceType
        fields = ("id", "name", "_detail")



from typing import Any

from django.core.exceptions import FieldError
from django.db.models import (
    CharField,
    Count,
    DecimalField,
    ExpressionWrapper,
    F,
    IntegerField,
    Q,
    QuerySet,
    Sum,
    Value,
)
from django.db.models.functions import Concat
from dynamic_preferences.registries import global_preferences_registry
from rest_framework import filters, status
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.currency.models import Currency, CurrencyFXRates
from wbcore.filters import DjangoFilterBackend

from wbaccounting.models import Invoice, submit_invoices_as_task
from wbaccounting.models.model_tasks import (
    approve_invoices_as_task,
    pay_invoices_as_task,
)
from wbaccounting.permissions import IsInvoiceAdmin
from wbaccounting.serializers import (
    ConsolidatedInvoiceSerializer,
    InvoiceModelSerializer,
    InvoiceRepresentationSerializer,
)
from wbaccounting.viewsets.buttons import (
    ConsolidatedInvoiceButtonConfig,
    InvoiceButtonConfig,
)
from wbaccounting.viewsets.display import (
    ConsolidatedInvoiceDisplayConfig,
    InvoiceDisplayConfig,
)
from wbaccounting.viewsets.endpoints import ConsolidatedInvoiceEndpointConfig
from wbaccounting.viewsets.titles import (
    ConsolidatedInvoiceTitleConfig,
    InvoiceTitleConfig,
)


class InvoiceRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = InvoiceRepresentationSerializer
    filter_backends = (filters.OrderingFilter,)
    queryset = Invoice.objects.all()
    ordering_fields = ["invoice_date"]
    ordering = ["invoice_date"]

    def get_queryset(self):
        return Invoice.objects.filter_for_user(self.request.user)  # type: ignore


class InvoiceModelViewSet(viewsets.ModelViewSet):
    serializer_class = InvoiceModelSerializer

    filterset_fields = {
        "counterparty": ["exact"],
        "status": ["exact"],
        "title": ["exact", "icontains"],
        "invoice_date": ["gte", "exact", "lte"],
        "invoice_type": ["exact"],
        "reference_date": ["gte", "exact", "lte"],
    }

    search_fields = ["counterparty__computed_str", "title", "invoice_type__name"]
    ordering_fields = (
        "invoice_date",
        "counterparty__computed_str",
        "invoice_type__name",
        "reference_date",
        "gross_value",
        "net_value",
    )
    ordering = (
        "-invoice_date",
        "id",
    )
    queryset = Invoice.objects.all()

    button_config_class = InvoiceButtonConfig
    display_config_class = InvoiceDisplayConfig
    title_config_class = InvoiceTitleConfig

    def get_queryset(self) -> QuerySet[Invoice]:
        invoices = Invoice.objects.filter_for_user(self.request.user)  # type: ignore

        if eai_id := self.kwargs.get("entry_accounting_information_id", None):
            invoices = invoices.filter(counterparty__entry_accounting_information__id=eai_id)

        return invoices.select_related(
            "counterparty",
            "invoice_currency",
            "invoice_type",
        )


class ConsolidatedInvoiceViewSet(viewsets.ViewSet):
    IDENTIFIER = "wbaccounting:consolidated-invoice"
    title_config_class = ConsolidatedInvoiceTitleConfig
    serializer_class = ConsolidatedInvoiceSerializer
    permission_classes = []
    display_config_class = ConsolidatedInvoiceDisplayConfig
    button_config_class = ConsolidatedInvoiceButtonConfig
    filterset_fields = {"status": ["exact"]}
    filter_backends = (DjangoFilterBackend,)
    endpoint_config_class = ConsolidatedInvoiceEndpointConfig
    queryset = Invoice.objects.none()

    def list(self, request):
        queryset = self.get_queryset()
        if serializer := self.get_serializer_class():
            serializer = serializer(queryset, many=True, context={"request": request, "view": self})
            return Response({"results": serializer.data})
        return Response({"results": {}})

    def get_lookup(self) -> dict:
        lookup = {}
        if (request := getattr(self, "request")) and (lookup_str := request.GET.get("lookup")):
            lookup = dict(map(lambda x: x.split(":"), lookup_str.split(",")))
        return lookup

    def get_lookup_mapping(self, currency_symbol: str = "") -> "list[tuple]":
        return [
            ("reference_date", "reference_date", "gross_value_default_currency", Value(currency_symbol)),
            ("invoice_type", "invoice_type__name", "gross_value_default_currency", Value(currency_symbol)),
            ("invoice_currency", "invoice_currency__symbol", "gross_value", F("invoice_currency__symbol")),
            ("counterparty", "counterparty__computed_str", "gross_value", F("invoice_currency__symbol")),
            ("id", "title", "gross_value", F("invoice_currency__symbol")),
            (
                "booking_entries",
                "booking_entries__title",
                "booking_entries__gross_value",
                F("booking_entries__currency__symbol"),
            ),
        ]

    def get_queryset(self):
        # We get the default currency id, as we potentially need it to convert some aggregates
        default_currency = Currency.objects.get(
            key=global_preferences_registry.manager()["currency__default_currency"]
        )

        # lookup is a GET parameter in the format of ?lookup=key1:param1,key2:param2
        # we convert it to a dictionairy for easier access
        lookup = self.get_lookup()

        # the number of items in the lookup dictionairy define the depth of the treeview, e.g
        # depth 0 means the first level, depth 1 the second, and so on
        depth = len(lookup.keys())

        # if there is a status GET parameter, we add it to make sure we only look at aggregated invoices
        # with that status
        status_filter = {}
        if status := self.request.GET.get("status"):
            status_filter["status"] = status

        # We filter the queryset with the lookup dictionairy and the status dictionairy
        # In case a wrong lookup was parsed in here a FieldError is raised and we return
        # an empty queryset
        try:
            queryset = Invoice.objects.filter_for_user(self.request.user).filter(**lookup, **status_filter)  # type: ignore
        except FieldError:
            return Invoice.objects.none()

        # Filter out all invoice that do not have a reference date.
        queryset = queryset.exclude(reference_date__isnull=True)

        # Depending on the depth and the default currency we get a tuple with the following elements
        # 0: reference to the field that will be used as an id
        # 1: reference to the name of the group
        # 2: reference to which field is used to sum up the invoices
        # 3: reference to which currency is used
        group_by = self.get_lookup_mapping(default_currency.symbol or "USD")[depth]

        # If the depth is smaller than 2, this means we have to convert the values to the default
        # currency FX Rate. This is because the first 2 levels (level 0 and 1) are group by date
        # and group by invoice type, which are currency agnostic. The third level (level 2) is by
        # currency, so we don't have to convert the invoice currency anymore
        if depth < 2:
            queryset = queryset.annotate(
                fx_rate_default_currency=CurrencyFXRates.get_fx_rates_subquery_for_two_currencies(
                    "reference_date", "invoice_currency", default_currency
                ),
                gross_value_default_currency=ExpressionWrapper(
                    F("gross_value") * F("fx_rate_default_currency"),
                    output_field=DecimalField(decimal_places=2, max_digits=15),
                ),
            )

        # if the id reference is not id (or level 4), we annotate the field as the id field
        # we cannot do this for level 4 (or when the id field is id) because django will
        # crash otherwise.
        if group_by[0] != "id":
            queryset = queryset.values(group_by[0]).annotate(id=F(group_by[0]))

        # We annotate all values received from the get_lookup_mapping
        queryset = queryset.annotate(
            currency_symbol=group_by[3],
            value=Sum(group_by[2]),
            group=F(group_by[1]),
            _group_key=Value(None, output_field=CharField(max_length=255)),
            depth=Value(depth, output_field=IntegerField()),
            num_draft=Count("pk", filter=Q(status="DRAFT")),
            num_submitted=Count("pk", filter=Q(status="SUBMITTED")),
            num_sent=Count("pk", filter=Q(status="SENT")),
            num_paid=Count("pk", filter=Q(status="PAID")),
        )

        # If the group_by id field is not booking entries (or not the last level), then we annotate a _group_key
        # We do this to give the frontend an understanding on how to open the next row. We don't do this on the
        # last level, to ensure that the last group cannot be further extended
        if group_by[0] != "booking_entries":
            queryset = queryset.annotate(
                _group_key=Concat(Value(group_by[0] + ":"), F(group_by[0]), output_field=CharField(max_length=255)),
            )

        return queryset.order_by("-reference_date").values(
            "id",
            "reference_date",
            "value",
            "currency_symbol",
            "_group_key",
            "group",
            "depth",
            "num_draft",
            "num_submitted",
            "num_sent",
            "num_paid",
        )

    def get_underlying_invoices(self, request: Request, pk: Any, source_status: Invoice.Status) -> QuerySet[Invoice]:
        """Returns all of the current row's underlying invoices filtered by `source_status`.

        Args:
            request: HttpRequest
            pk: Pk of the current row. Type varies accordingly.
            source_status (Invoice.Status): We need to filter by the state where a certain transition is allowed

        Returns:
            QuerySet(Invoice): A queryset of invoices
        """

        lookup = self.get_lookup()
        depth = len(lookup.keys())
        status_filter = request.GET.get("status", "")
        pk_filter = self.get_lookup_mapping()[depth][0]
        if depth < 5 and (not status_filter or status_filter == source_status.value):
            return Invoice.objects.filter(**lookup | {pk_filter: pk}, status=source_status)

        return Invoice.objects.none()

    @action(detail=True, methods=["PATCH"])
    def submit(self, request: Request, pk: Any):
        underlying_invoices = self.get_underlying_invoices(request, pk, Invoice.Status.DRAFT)
        submit_invoices_as_task.delay(underlying_invoices.values_list("id", flat=True))  # type: ignore
        return Response(status=status.HTTP_200_OK)

    @action(detail=True, methods=["PATCH"], permission_classes=[IsInvoiceAdmin])
    def approve(self, request: Request, pk: Any):
        underlying_invoices = self.get_underlying_invoices(request, pk, Invoice.Status.SUBMITTED)
        approve_invoices_as_task.delay(underlying_invoices.values_list("id", flat=True))  # type: ignore
        return Response(status=status.HTTP_200_OK)

    @action(detail=True, methods=["PATCH"], permission_classes=[IsInvoiceAdmin])
    def pay(self, request: Request, pk: Any):
        underlying_invoices = self.get_underlying_invoices(request, pk, Invoice.Status.SENT)
        pay_invoices_as_task.delay(underlying_invoices.values_list("id", flat=True))  # type: ignore
        return Response(status=status.HTTP_200_OK)



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

INVOICETYPE_MENUITEM = MenuItem(
    label="Invoice Types",
    endpoint="wbaccounting:invoicetype-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbaccounting.view_invoicetype"]
    ),
    add=MenuItem(
        label="Create Invoice Type",
        endpoint="wbaccounting:invoicetype-list",
        permission=ItemPermission(permissions=["wbaccounting.add_invoicetype"]),
    ),
)



from wbcore.metadata.configs.titles import TitleViewConfig


class InvoiceTypeTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Invoice Type"

    def get_list_title(self):
        return "Invoice Types"

    def get_create_title(self):
        return "New Invoice Type"



from .entry_accounting_information import EntryAccountingInformationButtonConfig
from .booking_entry import BookingEntryButtonConfig
from .invoice import InvoiceButtonConfig, ConsolidatedInvoiceButtonConfig



from .booking_entry import BookingEntryInline, BookingEntryModelAdmin
from .entry_accounting_information import EntryAccountingInformationModelAdmin
from .invoice_type import InvoiceTypeModelAdmin
from .invoice import InvoiceModelAdmin
from .transactions import TransactionModelAdmin



from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class TransactionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="booking_date", label=_("Booking Date"), width=150),
                dp.Field(key="value_date", label=_("Value Date"), width=150),
                dp.Field(key="bank_account", label=_("Bank Account"), width=450),
                # dp.Field(key="from_bank_account", label=_("Source"), width=300),
                # dp.Field(key="to_bank_account", label=_("Target"), width=300),
                dp.Field(key="value_local_ccy", label=_("Value (Local)"), width=150),
                dp.Field(key="value", label=_("Value"), width=150),
                dp.Field(key="description", label=_("Description"), width=450),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["booking_date", "value_date"],
                ["bank_account", "."],
                ["from_bank_account", "to_bank_account"],
                ["value", "currency"],
                ["value_local_ccy", "fx_rate"],
                ["description", "description"],
            ]
        )



from datetime import date

from django.db.models import Exists, OuterRef, QuerySet
from django.shortcuts import get_object_or_404
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.utils.date import get_date_interval_from_request

from wbaccounting.models import EntryAccountingInformation, Invoice
from wbaccounting.models.booking_entry import BookingEntry
from wbaccounting.serializers import (
    EntryAccountingInformationModelSerializer,
    EntryAccountingInformationRepresentationSerializer,
)
from wbaccounting.viewsets.buttons import EntryAccountingInformationButtonConfig
from wbaccounting.viewsets.display import EntryAccountingInformationDisplayConfig
from wbaccounting.viewsets.titles import EntryAccountingInformationTitleConfig

from ..permissions import CanGenerateBookingEntry, CanGenerateInvoice


class EntryAccountingInformationRepresentationViewSet(viewsets.RepresentationViewSet):
    search_fields = ("entry__computed_str",)
    queryset = EntryAccountingInformation.objects.all()
    serializer_class = EntryAccountingInformationRepresentationSerializer

    def get_queryset(self):
        return EntryAccountingInformation.objects.filter_for_user(self.request.user)


class EntryAccountingInformationModelViewSet(viewsets.ModelViewSet):
    IDENTIFIER = "wbaccounting:entryaccountinginformation"
    filterset_fields = {
        "entry": ["exact"],
        "tax_id": ["exact", "icontains"],
        "vat": ["gte", "exact", "lte"],
        "default_currency": ["gte", "exact", "lte"],
        "send_mail": ["exact"],
        "counterparty_is_private": ["exact"],
        "exempt_users": ["exact"],
    }

    serializer_class = EntryAccountingInformationModelSerializer
    queryset = EntryAccountingInformation.objects.all()

    search_fields = ["entry__computed_str"]
    ordering_fields = ["tax_id", "vat", "send_mail", "counterparty_is_private"]
    ordering = ["id"]

    display_config_class = EntryAccountingInformationDisplayConfig
    title_config_class = EntryAccountingInformationTitleConfig
    button_config_class = EntryAccountingInformationButtonConfig

    def get_queryset(self) -> QuerySet[EntryAccountingInformation]:
        return (
            EntryAccountingInformation.objects.filter_for_user(self.request.user)
            .select_related(
                "entry",
                "default_currency",
                "default_invoice_type",
            )
            .prefetch_related(
                "exempt_users",
                "email_to",
                "email_cc",
                "email_bcc",
            )
        )

    @action(detail=True, methods=["POST"], permission_classes=[CanGenerateBookingEntry])
    def generate_booking_entries(self, request, pk):
        eai = get_object_or_404(EntryAccountingInformation, id=pk)
        start, end = get_date_interval_from_request(request, request_type="POST")
        if start and end:
            eai.generate_booking_entries(start, end)  # type: ignore
            return Response(
                {"__notification": {"title": "Booking Entries are being generated in the background."}},
                status=status.HTTP_200_OK,
            )
        return Response({"status": "Missing start and end dates"}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=False, methods=["POST"], permission_classes=[CanGenerateBookingEntry])
    def generate_booking_entries_for_counterparties(self, request, pk=None):
        # TODO: this is not used yet, might want to clean this action
        start, end = get_date_interval_from_request(request, request_type="POST")
        if not start or not end:
            return Response({"status": "Missing start and end dates"}, status=status.HTTP_400_BAD_REQUEST)
        eais = EntryAccountingInformation.objects.filter(
            entry__id__in=request.POST.get("counterparties", "").split(",")
        )
        for eai in eais:
            eai.generate_booking_entries(start, end)  # type: ignore
        return Response(
            {"__notification": {"title": "Booking Entries are being generated in the background."}},
            status=status.HTTP_200_OK,
        )

    @action(detail=False, methods=["POST"], permission_classes=[CanGenerateInvoice])
    def generate_invoices_for_counterparties(self, request, pk=None):
        # TODO: this is not used yet, might want to clean this action
        eais = (
            EntryAccountingInformation.objects.filter(entry__id__in=request.POST.get("counterparties", "").split(","))
            .annotate(
                has_open_bookings=Exists(
                    BookingEntry.objects.filter(
                        counterparty=OuterRef("entry"),
                        invoice__isnull=True,
                        payment_date__isnull=True,
                    )
                )
            )
            .filter(has_open_bookings=True)
        )
        for eai in eais:
            Invoice.objects.create_for_counterparty(eai.entry, invoice_date=date.today())
        return Response(
            {"__notification": {"title": "Invoices are being generated in the background."}},
            status=status.HTTP_200_OK,
        )

    @action(detail=True, methods=["POST"], permission_classes=[CanGenerateInvoice])
    def invoice_booking_entries(self, request, pk):
        eai = get_object_or_404(EntryAccountingInformation, id=pk)
        Invoice.objects.create_for_counterparty(eai.entry, invoice_date=date.today())
        return Response(
            {"__notification": {"title": "Booking Entries are being generated in the background."}},
            status=status.HTTP_200_OK,
        )



from django.contrib import admin

from wbaccounting.models import EntryAccountingInformation


@admin.register(EntryAccountingInformation)
class EntryAccountingInformationModelAdmin(admin.ModelAdmin):
    list_display = ("entry", "send_mail", "counterparty_is_private")
    autocomplete_fields = ["default_currency", "entry", "email_to", "email_cc", "email_bcc", "exempt_users"]

    search_fields = ["entry__computed_str"]



from decimal import Decimal

import factory
import pytest
from django.apps import apps
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from wbaccounting.factories import (
    BookingEntryFactory,
    EntryAccountingInformationFactory,
    InvoiceFactory,
    InvoiceTypeFactory,
    LocalCurrencyTransactionFactory,
    TransactionFactory,
)
from wbcore.contrib.authentication.factories import (
    SuperUserFactory,
    UserActivityFactory,
    UserFactory,
)
from wbcore.contrib.currency.factories import CurrencyFactory, CurrencyFXRatesFactory
from wbcore.contrib.directory.factories import (
    BankingContactFactory,
    EntryFactory,
    PersonFactory,
)
from wbcore.contrib.geography.tests.signals import app_pre_migration
from wbcore.tests.conftest import *

register(BookingEntryFactory)
register(InvoiceFactory)
register(InvoiceTypeFactory)
register(EntryAccountingInformationFactory)

register(TransactionFactory)
register(TransactionFactory, "transaction_no_value_date", value_date=None)
register(
    TransactionFactory,
    "transaction_fx",
    fx_rate=factory.Faker("pydecimal", min_value=Decimal(0.1), max_value=Decimal(0.9)),
)
register(LocalCurrencyTransactionFactory, "transaction_local_ccy")
register(
    LocalCurrencyTransactionFactory,
    "transaction_local_ccy_fx",
    fx_rate=factory.Faker("pydecimal", min_value=Decimal(0.1), max_value=Decimal(0.9)),
)


register(CurrencyFactory)
register(CurrencyFXRatesFactory)

register(EntryFactory)
register(UserFactory)
register(PersonFactory)
register(SuperUserFactory)
register(UserActivityFactory)
register(BankingContactFactory)

pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("wbaccounting"))


@pytest.fixture
def booking_entries(request, booking_entry_factory):
    return [booking_entry_factory.create() for _ in range(request.param)]



from wbcore.metadata.configs.titles import TitleViewConfig


class BookingEntryTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Booking: {{title}}"

    def get_list_title(self):
        return "Booking"

    def get_create_title(self):
        return "New Booking"



from .booking_entry import BookingEntryFactory
from .entry_accounting_information import (
    CompanyAccountingFactory,
    EntryAccountingInformationFactory,
)
from .invoice import (
    InvoiceFactory,
    InvoiceTypeFactory,
)
from .transactions import TransactionFactory, LocalCurrencyTransactionFactory



from .booking_entry import (
    BookingEntryModelViewSet,
    BookingEntryRepresentationViewSet,
)
from .entry_accounting_information import (
    EntryAccountingInformationModelViewSet,
    EntryAccountingInformationRepresentationViewSet,
)
from .invoice_type import InvoiceTypeModelViewSet, InvoiceTypeRepresentationViewSet
from .invoice import InvoiceModelViewSet, ConsolidatedInvoiceViewSet, InvoiceRepresentationViewSet
from .transactions import TransactionModelViewSet, TransactionRepresentationViewSet
from .cashflows import FutureCashFlowPandasAPIViewSet, FutureCashFlowTransactionsPandasAPIViewSet



from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class InvoiceTypeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Name"),
                dp.Field(key="processor", label="Processor"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["name", "processor"]])



from wbcore.metadata.configs.titles import TitleViewConfig


class EntryAccountingInformationTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Counterparty: {{ _entry.computed_str }}"

    def get_list_title(self):
        return "Counterparties"

    def get_create_title(self):
        return "New Counterparty"



import abc
from contextlib import suppress
from datetime import date
from importlib import import_module
from typing import Callable, Iterable

from django.conf import settings
from django.db.models import QuerySet
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.directory.models import Entry

from wbaccounting.models.booking_entry import BookingEntry

GENERATE_BOOKING_ENTRIES = Callable[[date, date, Entry], list[BookingEntry]]


def register_generator(name: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        setattr(func, "_is_booking_entry_generator", True)
        setattr(func, "_booking_entry_generator_name", name)
        return func

    return decorator


def get_all_booking_entry_choices() -> Iterable[tuple[str, str]]:
    for app in settings.INSTALLED_APPS:
        with suppress(ModuleNotFoundError):
            import_module(f"{app}.generators")

    for subclass in AbstractBookingEntryGenerator.__subclasses__():
        yield f"{subclass.__module__}.{subclass.__name__}", subclass.TITLE


class AbstractBookingEntryGenerator(abc.ABC):
    """
    An abstract base class designed to define a template for generating and managing
    booking entries within a financial or accounting system.

    This class outlines the necessary operations for generating a series of booking
    entries based on a given date range and counterparty, as well as merging backlinks
    for the generated booking entries.

    Attributes:
        TITLE (str): A class-level attribute that provides a title or a descriptive name for the
                     generator implementation. This should be overridden in concrete subclasses
                     to provide a specific title.
    """

    TITLE = ""

    @staticmethod
    @abc.abstractmethod
    def generate_booking_entries(from_date: date, to_date: date, counterparty: Entry) -> Iterable[BookingEntry]:
        """
        Generates a sequence of booking entries for a specified date range and counterparty.

        This abstract method must be implemented by subclasses to provide the logic for
        generating booking entries based on the provided criteria.

        Args:
            from_date (date): The start date of the period for which booking entries are to be generated.
            to_date (date): The end date of the period for which booking entries are to be generated.
            counterparty (Entry): The counterparty associated with the booking entries to be generated.

        Returns:
            Iterable[BookingEntry]: A sequence of BookingEntry instances generated for the specified
                                    criteria.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.
        """
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def merge_backlinks(booking_entries: QuerySet[BookingEntry]) -> dict:
        """
        Merges backlinks for a collection of booking entries, to consolidate
        related entries to insert a single backling in an invoice.

        This abstract method should be implemented by subclasses to provide specific
        logic for merging or updating backlinks of booking entries.

        Args:
            booking_entries (Iterable[BookingEntry]): A sequence of BookingEntry instances
                                                     to be merged or updated.

        Returns:
            dict: A dictionary or other structured data indicating the results of the
                  merge operation, specific to the implementation.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.
        """
        raise NotImplementedError()


def generate_booking_entries(
    _class: type[AbstractBookingEntryGenerator], from_date: date, to_date: date, counterparty: Entry
):
    booking_entries = _class.generate_booking_entries(from_date, to_date, counterparty)
    for booking_entry in booking_entries:
        booking_entry.save()


class TestGenerator(AbstractBookingEntryGenerator):
    TITLE = "Test Generator"

    @staticmethod
    def generate_booking_entries(from_date: date, to_date: date, counterparty: Entry) -> Iterable[BookingEntry]:
        yield BookingEntry(
            title="Test Booking Entry",
            counterparty=counterparty,
            booking_date=date.today(),
            reference_date=date.today(),
            net_value=100,
            currency=Currency.objects.first(),
        )

    @staticmethod
    def merge_backlinks(booking_entries: Iterable[BookingEntry]) -> dict:
        return {}



import factory

from wbaccounting.models import Invoice, InvoiceType


class InvoiceFactory(factory.django.DjangoModelFactory):
    class Meta:  # type: ignore
        model = Invoice

    title = factory.Faker("text", max_nb_chars=64)
    invoice_date = factory.Faker("date_object")
    reference_date = factory.Faker("date_object")
    invoice_currency = factory.SubFactory("wbcore.contrib.currency.factories.CurrencyUSDFactory")

    counterparty = factory.SubFactory("wbcore.contrib.directory.factories.EntryFactory")
    invoice_type = factory.SubFactory("wbaccounting.factories.InvoiceTypeFactory")

    text_above = factory.Faker("text")
    text_below = factory.Faker("text")

    # @factory.post_generation
    # def invoice_document(self, create, extracted, **kwargs):
    #     self.refresh_invoice_document(override_status=True)

    # is_counterparty_invoice

    # @classmethod
    # def _create(cls, model_class, *args, **kwargs):
    #     company = CompanyFactory()
    #     signee = PersonSignatureFactory()
    #     global_preferences_registry.manager()["wbaccounting__invoice_company"] = company.id
    #     global_preferences_registry.manager()["wbaccounting__invoice_signers"] = Person.objects.filter(id=signee.id)
    #     """Override the default ``_create`` with our custom call."""
    #     manager = cls._get_manager(model_class)
    #     # The default would use ``manager.create(*args, **kwargs)``
    #     return manager.create(*args, **kwargs)


class InvoiceTypeFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence(lambda n: f"Invoice Type {n}")
    processor = factory.Faker("text", max_nb_chars=64)

    class Meta:  # type: ignore
        model = InvoiceType



# Generated by Django 5.0.6 on 2024-06-05 08:07

from decimal import Decimal

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("currency", "0001_initial"),
        ("directory", "0007_alter_bankingcontact_options"),
        ("io", "0007_alter_exportsource_query_params"),
        ("wbaccounting", "0010_alter_bookingentry_options"),
    ]

    operations = [
        migrations.CreateModel(
            name="Transaction",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("booking_date", models.DateField(verbose_name="Booking Date")),
                ("value_date", models.DateField(verbose_name="Value Date")),
                (
                    "fx_rate",
                    models.DecimalField(decimal_places=4, default=Decimal("1"), max_digits=10, verbose_name="FX Rate"),
                ),
                (
                    "value_local_ccy",
                    models.DecimalField(
                        blank=True, decimal_places=2, max_digits=19, null=True, verbose_name="Value (Local Currency)"
                    ),
                ),
                (
                    "value",
                    models.DecimalField(blank=True, decimal_places=2, max_digits=19, null=True, verbose_name="Value"),
                ),
                ("description", models.TextField(default="")),
                (
                    "prenotification",
                    models.BooleanField(
                        default=False,
                        help_text="This field indicates that this transaction will happen sometime in the future.",
                        verbose_name="Prenotification",
                    ),
                ),
                ("_hash", models.CharField(blank=True, max_length=64, null=True)),
                (
                    "bank_account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="wbaccounting_transactions",
                        to="directory.bankingcontact",
                        verbose_name="Linked Bank Account",
                    ),
                ),
                (
                    "currency",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="currency.currency",
                        verbose_name="Currency",
                    ),
                ),
                (
                    "from_bank_account",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="directory.bankingcontact",
                        verbose_name="Source Bank Account",
                    ),
                ),
                (
                    "import_source",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="io.importsource"
                    ),
                ),
                (
                    "to_bank_account",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="directory.bankingcontact",
                        verbose_name="Target Bank Account",
                    ),
                ),
            ],
            options={
                "verbose_name": "Transaction",
                "verbose_name_plural": "Transactions",
                "default_related_name": "transactions",
            },
        ),
    ]



from decimal import Decimal

from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from wbcore.contrib.authentication.models import User
from wbcore.models import WBModel


class BookingEntryDefaultQuerySet(models.QuerySet):
    def filter_for_user(self, user: User) -> models.QuerySet:
        """
        Filters booking entries based on if the current user can see the invoice.

        Args:
            user (User): The user for whom booking entries need to be filtered.

        Returns:
            QuerySet: A filtered queryset.
        """

        # Superuser and users with the admin permission can see all booking entries
        if user.is_superuser or user.has_perm("wbaccounting.administrate_invoice"):
            return self

        # If the user doesn't have the view permission, nothing can be seen
        if not user.has_perm("wbaccounting.view_bookingentry"):
            return self.none()

        # The user can see all booking entries where the counterparty is not private
        # or where the user is part of the exempt users list
        return self.filter(
            models.Q(counterparty__entry_accounting_information__counterparty_is_private=False)
            | models.Q(counterparty__entry_accounting_information__exempt_users=user)
        )


class BookingEntry(WBModel):
    class Meta:
        verbose_name = "Booking"
        verbose_name_plural = "Bookings"

        permissions = (
            (
                "can_generate_booking_entries",
                "Can Generate Bookings",
            ),
        )

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        # Reference date defaults to booking date if not specified
        if not self.reference_date:
            self.reference_date = self.booking_date

        # If net value is specified, then gross value is determined based on VAT
        if self.net_value:
            self.gross_value = self.net_value + (self.net_value * self.vat)
        # If net value is not specified, but gross value is, then net value is determined based on VAT
        elif self.gross_value:
            self.net_value = self.gross_value / (Decimal(1 + self.vat))
        # If neither is defined the both are 0
        else:
            self.net_value, self.gross_value = 0, 0

        # If an invoice is attached we need to compute net and gross value in the invoice currency
        if self.invoice:
            # We get the fx rate (If it is the same currency, then it will be 1)
            self.invoice_fx_rate = self.currency.convert(self.booking_date, self.invoice.invoice_currency)
            self.invoice_net_value = self.net_value * self.invoice_fx_rate
            self.invoice_gross_value = self.gross_value * self.invoice_fx_rate

        super().save(*args, **kwargs)

    objects = BookingEntryDefaultQuerySet.as_manager()

    # The title of the booking entry which describes what was booked
    title = models.CharField(max_length=255, verbose_name="Title")

    # The date when this booking entry is booked. Most likely the current date
    booking_date = models.DateField(verbose_name="Booking Date")

    # The date when this booking entry has to be paid / received. If None, this date is not important.
    due_date = models.DateField(null=True, blank=True, verbose_name="Due Date")

    # The date when this booking entry was paid / received. If None, this booking entry is still "open"
    payment_date = models.DateField(null=True, blank=True, verbose_name="Payment Date")

    # The reference date represents a date to which accounting period a booking entry belongs
    reference_date = models.DateField(verbose_name="Reference Date", null=True, blank=True)

    # Either Gross or Net has to be specified. If both are specifying then net is preferred.
    # If only one is specified the other one is determined based on the given VAT. The default
    # VAT is 0%. If the value of net/gross is negative, then money is paid.
    gross_value = models.DecimalField(
        max_digits=16, decimal_places=4, null=True, blank=True, verbose_name="Gross Value"
    )
    net_value = models.DecimalField(max_digits=16, decimal_places=4, null=True, blank=True, verbose_name="Net Value")
    invoice_gross_value = models.DecimalField(
        max_digits=16, decimal_places=4, null=True, blank=True, verbose_name="Invoice Gross Value"
    )
    invoice_net_value = models.DecimalField(
        max_digits=16, decimal_places=4, null=True, blank=True, verbose_name="Invoice Net Value"
    )
    invoice_fx_rate = models.DecimalField(
        max_digits=20, decimal_places=6, null=True, blank=True, verbose_name="Invoice FX Rate"
    )
    vat = models.DecimalField(max_digits=4, decimal_places=4, default=Decimal(0), verbose_name="VAT")

    # The currency of the booking entry
    currency = models.ForeignKey(
        "currency.Currency", related_name="booking_entries", on_delete=models.PROTECT, verbose_name="Currency"
    )

    invoice = models.ForeignKey(
        "Invoice",
        related_name="booking_entries",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Invoice",
    )

    # The entry who has to pay / receive money
    counterparty = models.ForeignKey(
        "directory.Entry", related_name="booking_entries", on_delete=models.PROTECT, verbose_name="Counterparty"
    )

    # The backlinks stores information and the destination where to find the underlying data
    # The backlinks is rendered in lists and instances as a button
    backlinks = models.JSONField(null=True, blank=True)

    # Extra parameters for rendering on an invoice
    parameters = models.JSONField(null=True, blank=True)

    # Generator that produced this Booking Entry
    generator = models.CharField(max_length=256, null=True, blank=True)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbaccounting:bookingentry"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbaccounting:bookingentryrepresentation-list"


@receiver(post_save, sender=BookingEntry)
def booking_entry_changed(sender, instance: BookingEntry, created: bool, raw: bool, **kwargs):
    if not raw and (invoice := instance.invoice):
        invoice.save()


@receiver(post_delete, sender=BookingEntry)
def booking_entry_deleted(sender, instance, **kwargs):
    if instance.invoice:
        instance.invoice.save()



from datetime import date, datetime
from typing import TYPE_CHECKING, Any

import pandas as pd
from django.db.models import DateField, F, Sum, Value
from wbcore.pandas import fields as pf
from wbcore.pandas.views import PandasAPIViewSet

from wbaccounting.models import Transaction
from wbaccounting.viewsets.display import FutureCashFlowDisplayConfig

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from django.db.models.query import ValuesQuerySet
    from rest_framework.request import Request


class FutureCashFlowPandasAPIViewSetMixin(PandasAPIViewSet):
    DATE_COL_START = 0
    BOLD = False

    queryset = Transaction.objects.none()
    display_config_class = FutureCashFlowDisplayConfig

    def get_pandas_fields(self, request) -> pf.PandasFields:
        fields = self.get_dataframe(request, self.get_queryset()).columns[self.DATE_COL_START :]
        return pf.PandasFields(
            fields=[
                pf.PKField(key="bank_account__id", label="ID"),
                pf.CharField(key="_group_key", label="GROUPKEY"),
                pf.CharField(key="bank_account__iban", label="IBAN"),
                pf.CharField(key="bank_account__currency__symbol", label="Currency"),
                *[
                    pf.FloatField(
                        key=field,
                        label=datetime.strptime(field, "%Y-%m-%d").strftime("%d.%m.%Y"),
                        display_mode=pf.DisplayMode.SHORTENED,
                    )
                    for field in fields
                ],
            ]
        )


class FutureCashFlowPandasAPIViewSet(FutureCashFlowPandasAPIViewSetMixin):
    DATE_COL_START = 4
    BOLD = True

    def get_queryset(self) -> "ValuesQuerySet[Transaction, Any]":
        base_queryset = Transaction.objects.filter_for_user(self.request.user)
        if banking_contact_ids := self.request.GET.get("banking_contact"):
            base_queryset = base_queryset.filter(bank_account__id__in=banking_contact_ids.split(","))

        past_transactions = (
            base_queryset.filter(prenotification=False)
            .values("bank_account__id")
            .annotate(
                value_date=Value(date.today(), output_field=DateField()),
                balance=Sum("value"),
                _group_key=F("bank_account__id"),
            )
            .values(
                "bank_account__id",
                "bank_account__iban",
                "bank_account__currency__symbol",
                "value_date",
                "balance",
                "_group_key",
            )
        )

        future_transactions = (
            base_queryset.filter(prenotification=True, value_date__gt=date.today())
            .values("bank_account__id", "value_date")
            .annotate(balance=Sum("value"), _group_key=F("bank_account__id"))
            .values(
                "bank_account__id",
                "bank_account__iban",
                "bank_account__currency__symbol",
                "value_date",
                "balance",
                "_group_key",
            )
        )

        return past_transactions.union(future_transactions)

    def get_dataframe(
        self, request: "Request", queryset: "ValuesQuerySet[Transaction, Any]", **kwargs
    ) -> pd.DataFrame:
        if not queryset.exists():
            return pd.DataFrame()
        df = (
            pd.DataFrame(queryset)
            .pivot_table(
                index=["bank_account__id", "bank_account__iban", "bank_account__currency__symbol", "_group_key"],
                columns=["value_date"],
                values="balance",
            )
            .astype(float)
            .fillna(0)
            .cumsum(axis=1)
        )
        df.columns = df.columns.astype(str)
        return df.reset_index()


class FutureCashFlowTransactionsPandasAPIViewSet(FutureCashFlowPandasAPIViewSetMixin):
    DATE_COL_START = 2

    def get_queryset(self):
        queryset = Transaction.objects.filter(prenotification=True, value_date__gt=date.today())
        if bank_account := self.request.GET.get("bank_account", None):
            return queryset.filter(bank_account_id=bank_account)
        return queryset

    def get_dataframe(self, request: "Request", queryset: "QuerySet[Transaction]", **kwargs) -> pd.DataFrame:
        df = pd.DataFrame(
            queryset.values_list("description", "value", "value_date"), columns=["description", "value", "value_date"]
        ).pivot_table(index=["description"], columns=["value_date"], values="value")
        df.columns = df.columns.astype(str)
        df = df.reset_index()
        df = df.rename(columns={"description": "bank_account__iban"})
        df.insert(0, "bank_account__id", df["bank_account__iban"])
        return df



from .booking_entry import BookingEntryTitleConfig
from .entry_accounting_information import EntryAccountingInformationTitleConfig
from .invoice import InvoiceTitleConfig, ConsolidatedInvoiceTitleConfig
from .invoice_type import InvoiceTypeTitleConfig



import pytest
from wbaccounting.models import Transaction


@pytest.mark.django_db
class TestTransaction:
    def test_str(self, transaction: Transaction):
        assert str(transaction) == f"{transaction.booking_date:%d.%m.%Y}: {transaction.value:.2f}"

    @pytest.mark.parametrize(
        "method,return_value",
        [
            ("get_endpoint_basename", "wbaccounting:transaction"),
            ("get_representation_value_key", "id"),
            ("get_representation_label_key", "{{booking_date}}: {{value}}"),
            ("get_representation_endpoint", "wbaccounting:transactionrepresentation-list"),
        ],
    )
    def test_wbmodel_methods(self, method: str, return_value: str):
        assert getattr(Transaction, method)() == return_value

    def test_no_value_date(self, transaction_no_value_date: Transaction):
        assert transaction_no_value_date.value_date == transaction_no_value_date.booking_date

    def test_value(self, transaction: Transaction):
        assert pytest.approx(transaction.value_local_ccy) == transaction.value

    def test_value_different_fx_rate(self, transaction_fx: Transaction):
        tfx = transaction_fx
        assert tfx.value is not None
        assert pytest.approx(tfx.value / tfx.fx_rate) == pytest.approx(tfx.value_local_ccy)

    def test_local_ccy(self, transaction_local_ccy: Transaction):
        tlc = transaction_local_ccy
        assert pytest.approx(tlc.value) == tlc.value_local_ccy

    def test_local_ccy_different_fx_rate(self, transaction_local_ccy_fx: Transaction):
        tlcf = transaction_local_ccy_fx
        assert tlcf.value_local_ccy is not None
        assert pytest.approx(tlcf.value) == (tlcf.value_local_ccy * tlcf.fx_rate)



from django.conf import settings
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import (
    IntegerPreference,
    LongStringPreference,
    ModelMultipleChoicePreference,
    StringPreference,
)
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.directory.models import Person

accounting_section = Section("wbaccounting")


def format_invoice_number(number):
    global_preferences = global_preferences_registry.manager()
    ts = global_preferences["wbaccounting__invoice_thousand_seperator"]
    ds = global_preferences["wbaccounting__invoice_decimal_seperator"]
    if not number:
        number = 0
    s = "{:,.2f}".format(number)
    s = s.replace(",", "//")
    s = s.replace(".", "\\")

    return s.replace("//", ts).replace("\\", ds)


@global_preferences_registry.register
class DefaultEntryAccountingInformationCurrency(StringPreference):
    section = accounting_section
    name = "default_entry_account_information_currency_key"
    default = "CHF"

    verbose_name = "The currency key used for the default currency of newly created entry accounting informations."

    def validate(self, value):
        if not Currency.objects.filter(key=value).exists():
            return ValueError("The specified currency key is not valid")


@global_preferences_registry.register
class ExternalEmailAddress(StringPreference):
    section = accounting_section
    name = "external_email_address"
    default = ""

    verbose_name = "External Email Address"
    help_text = "The Email Address used to send Invoices to the external party."


@global_preferences_registry.register
class InvoiceEmailBody(LongStringPreference):
    section = accounting_section
    name = "invoice_email_body"
    default = ""

    verbose_name = "The default E-Mail Body used for emailing Invoices"


@global_preferences_registry.register
class InvoiceThousandSeperatorPreference(StringPreference):
    section = accounting_section
    name = "invoice_thousand_seperator"
    default = ","

    verbose_name = "Invoice Thousand Seperator"
    help_text = "Thousand Seperator for Invoices"


@global_preferences_registry.register
class InvoiceDecimalSeperatorPreference(StringPreference):
    section = accounting_section
    name = "invoice_decimal_seperator"
    default = "."

    verbose_name = "Invoice Decimal Seperator"
    help_text = "Decimal Seperator for Invoices"


@global_preferences_registry.register
class InvoiceSignerPreference(ModelMultipleChoicePreference):
    section = accounting_section
    name = "invoice_signers"
    queryset = Person.objects.all()
    default = None
    verbose_name = "Invoice Signers"


@global_preferences_registry.register
class InvoiceCompanyPreference(IntegerPreference):
    section = accounting_section
    name = "invoice_company"
    default = 0

    verbose_name = "Invoice Company"
    help_text = "The PK of the company who issues the invoices"


@global_preferences_registry.register
class DefaultFromEmailAddressPreference(StringPreference):
    section = accounting_section
    name = "default_from_email_address"
    default = settings.DEFAULT_FROM_EMAIL

    verbose_name = "The default from email address"
    help_text = "The default from email address used to send invoice"



# Generated by Django 4.1.7 on 2023-04-19 05:38

import django.db.models.deletion
import django_fsm
import wbaccounting.models
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("dynamic_preferences", "0006_auto_20191001_2236"),
        ("directory", "0001_initial"),
        ("currency", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Invoice",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "status",
                    django_fsm.FSMField(
                        choices=[
                            ("DRAFT", "Draft"),
                            ("PENDING", "Pending"),
                            ("EXTERNAL", "External"),
                            ("SENT", "Sent"),
                            ("PAID", "Paid"),
                        ],
                        default="DRAFT",
                        max_length=50,
                        verbose_name="Status",
                    ),
                ),
                ("resolved", models.BooleanField(default=False, verbose_name="Resolved")),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                ("invoice_date", models.DateField(verbose_name="Date")),
                ("text_above", models.TextField(blank=True, null=True)),
                ("text_below", models.TextField(blank=True, null=True)),
                ("is_counterparty_invoice", models.BooleanField(default=False)),
                (
                    "invoice_currency",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="invoices",
                        to="currency.currency",
                        verbose_name="Curency",
                    ),
                ),
            ],
            options={
                "verbose_name": "Invoice",
                "verbose_name_plural": "Invoices",
                "permissions": (("can_generate_invoice", "Can Generate Invoice"),),
            },
        ),
        migrations.CreateModel(
            name="BookingEntry",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("resolved", models.BooleanField(default=False, verbose_name="Resolved")),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                ("booking_date", models.DateField(verbose_name="Booking Date")),
                ("accrual_date", models.DateField(blank=True, null=True, verbose_name="Real Date")),
                ("from_date", models.DateField(blank=True, null=True, verbose_name="From")),
                ("to_date", models.DateField(blank=True, null=True, verbose_name="To")),
                (
                    "gross_value",
                    models.DecimalField(
                        blank=True, decimal_places=4, max_digits=16, null=True, verbose_name="Gross Value"
                    ),
                ),
                ("vat", models.DecimalField(decimal_places=4, default=0.0, max_digits=4, verbose_name="VAT")),
                (
                    "net_value",
                    models.DecimalField(
                        blank=True, decimal_places=4, max_digits=16, null=True, verbose_name="Net Value"
                    ),
                ),
                ("resolved_date", models.DateField(blank=True, null=True, verbose_name="Resolved Date")),
                (
                    "invoice",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="booking_entries",
                        to="wbaccounting.invoice",
                        verbose_name="Invoice",
                    ),
                ),
                ("related_data", models.JSONField(default=dict, verbose_name="Related Data")),
                (
                    "calculated_value",
                    models.DecimalField(
                        blank=True, decimal_places=4, max_digits=16, null=True, verbose_name="Final Value"
                    ),
                ),
                (
                    "main_booking_entry",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="appended_booking_entries",
                        to="wbaccounting.bookingentry",
                        verbose_name="Main Booking Entry",
                    ),
                ),
                (
                    "currency",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="booking_entries",
                        to="currency.currency",
                        verbose_name="Currency",
                    ),
                ),
            ],
            options={
                "verbose_name": "Booking Entry",
                "verbose_name_plural": "Booking Entries",
            },
        ),
        migrations.CreateModel(
            name="EntryAccountingInformation",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("tax_id", models.CharField(blank=True, max_length=512, null=True, verbose_name="Tax ID")),
                ("vat", models.FloatField(blank=True, null=True, verbose_name="VAT")),
                ("social_charges", models.FloatField(blank=True, null=True, verbose_name="Social Charges")),
                (
                    "label_mapping",
                    models.JSONField(default=dict, verbose_name="Label Mapping"),
                ),
                ("email_subject", models.TextField(default="{{invoice.title}}", verbose_name="Subject")),
                ("email_body", models.TextField(default="", verbose_name="Body")),
                (
                    "default_currency",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="entry_accounting_informations",
                        to="currency.currency",
                        verbose_name="Default Currency",
                    ),
                ),
            ],
            options={
                "verbose_name": "Entry Accounting Information",
                "verbose_name_plural": "Entry Accounting Information",
            },
        ),
        migrations.AddField(
            model_name="bookingentry",
            name="counterparty",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="booking_entries",
                to="directory.entry",
                verbose_name="Counterparty",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="email_bcc",
            field=models.ManyToManyField(
                blank=True,
                related_name="entry_accounting_informations_bcc",
                to="directory.emailcontact",
                verbose_name="BCC",
            ),
        ),
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="email_cc",
            field=models.ManyToManyField(
                blank=True,
                related_name="entry_accounting_informations_cc",
                to="directory.emailcontact",
                verbose_name="CC",
            ),
        ),
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="email_to",
            field=models.ManyToManyField(
                blank=True,
                related_name="entry_accounting_informations_to",
                to="directory.emailcontact",
                verbose_name="To",
            ),
        ),
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="entry",
            field=models.OneToOneField(
                default=None,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="entry_accounting_information",
                to="directory.entry",
                verbose_name="Entry",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="invoice",
            name="counterparty",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="accounting_invoices",
                to="directory.entry",
                verbose_name="Counterparty",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="bookingentry",
            name="counterparty",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="booking_entries",
                to="directory.entry",
                verbose_name="Counterparty",
            ),
        ),
        migrations.AlterField(
            model_name="entryaccountinginformation",
            name="email_bcc",
            field=models.ManyToManyField(
                blank=True,
                related_name="entry_accounting_informations_bcc",
                to="directory.emailcontact",
                verbose_name="BCC",
            ),
        ),
        migrations.AlterField(
            model_name="entryaccountinginformation",
            name="email_cc",
            field=models.ManyToManyField(
                blank=True,
                related_name="entry_accounting_informations_cc",
                to="directory.emailcontact",
                verbose_name="CC",
            ),
        ),
        migrations.AlterField(
            model_name="entryaccountinginformation",
            name="email_to",
            field=models.ManyToManyField(
                blank=True,
                related_name="entry_accounting_informations_to",
                to="directory.emailcontact",
                verbose_name="To",
            ),
        ),
        migrations.AlterField(
            model_name="entryaccountinginformation",
            name="entry",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="entry_accounting_information",
                to="directory.entry",
                verbose_name="Entry",
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="counterparty",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="accounting_invoices",
                to="directory.entry",
                verbose_name="Counterparty",
            ),
        ),
    ]



import factory
import pytest
from rest_framework.reverse import reverse
from wbaccounting.models import InvoiceType
from wbaccounting.serializers import (
    InvoiceTypeModelSerializer,
    InvoiceTypeRepresentationSerializer,
)


@pytest.mark.django_db
class TestInvoiceTypeModelSerializer:
    def test_serialize(self, invoice_type: InvoiceType):
        serializer = InvoiceTypeModelSerializer(invoice_type)
        assert serializer.data == {
            "id": invoice_type.pk,
            "name": invoice_type.name,
            "processor": invoice_type.processor,
        }

    def test_deserialize(self, invoice_type_factory):
        data = factory.build(dict, FACTORY_CLASS=invoice_type_factory)
        serializer = InvoiceTypeModelSerializer(data=data)
        assert serializer.is_valid()


@pytest.mark.django_db
class TestInvoiceTypeRepresentationSerializer:
    def test_serialize(self, invoice_type: InvoiceType):
        serializer = InvoiceTypeRepresentationSerializer(invoice_type)
        assert serializer.data == {
            "id": invoice_type.pk,
            "name": invoice_type.name,
            "_detail": reverse("wbaccounting:invoicetype-detail", args=[invoice_type.pk]),
        }



# Generated by Django 4.1.10 on 2023-08-30 11:20

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbaccounting", "0006_alter_invoice_status"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="invoice",
            options={
                "permissions": (
                    ("can_generate_invoice", "Can Generate Invoice"),
                    ("administrate_invoice", "Can administer Invoice"),
                ),
                "verbose_name": "Invoice",
                "verbose_name_plural": "Invoices",
            },
        ),
    ]



# Generated by Django 4.1.9 on 2023-05-16 14:46

import django_fsm
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbaccounting", "0001_initial_squashed_squashed_0005_alter_bookingentry_counterparty_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="status",
            field=django_fsm.FSMField(
                choices=[
                    ("DRAFT", "Draft"),
                    ("PENDING", "Pending"),
                    ("EXTERNAL", "External"),
                    ("SENT", "Sent"),
                    ("PAID", "Paid"),
                    ("CANCELLED", "Cancelled"),
                ],
                default="DRAFT",
                max_length=50,
                verbose_name="Status",
            ),
        ),
    ]



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

CASHFLOW_MENUITEM = MenuItem(
    label="Cashflow",
    endpoint="wbaccounting:futurecashflow-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbaccounting.view_transaction"]
    ),
)



from django.db import models
from wbcore.models import WBModel


class InvoiceType(WBModel):
    name = models.CharField(max_length=100, unique=True, verbose_name="Name")
    processor = models.CharField(max_length=128, null=True, blank=True, verbose_name="Processor")

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbaccounting:invoicetype"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbaccounting:invoicetyperepresentation-list"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"

    class Meta:  # type: ignore
        verbose_name = "Invoice Type"
        verbose_name_plural = "Invoice Types"
